import numpy as np
from sklearn.cross_decomposition import PLSRegression
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import mean_squared_error
import scipy
# https://towardsdatascience.com/feature-reduction-using-genetic-algorithm-with-python-403a5f4ef0c1

def reduce_features(solution, features):
    selected_elements_indices = np.where(solution == 1)[0]
    reduced_features = features[:, selected_elements_indices]
    return reduced_features


def classification_accuracy(labels, predictions):
    return rsquared(labels, predictions)[0]

def rsquared(x, y):
    """ Return R^2 where x and y are array-like."""

    slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(x[:,0], y[:, 0])
    return [r_value**2, p_value, std_err]


def cal_pop_fitness(pop, features, labels, train_indices, test_indices):
    accuracies = np.zeros(pop.shape[0])
    idx = 0
    vaScore = np.zeros(features.shape[1])
    print(vaScore.shape)
    for curr_solution in pop:
        reduced_features = reduce_features(curr_solution, features)
        train_data = reduced_features[train_indices, :]
        test_data = reduced_features[test_indices, :]

        train_labels = labels[train_indices]
        test_labels = labels[test_indices]

        parameters = {"n_components" : np.arange(1,train_data.shape[1])}
        clf = GridSearchCV(PLSRegression(), parameters, cv=5, n_jobs=8, scoring='neg_root_mean_squared_error', verbose=0, refit=True)
        clf.fit(train_data, train_labels)
        bestMod = clf.best_estimator_
        accuracies[idx] = -mean_squared_error(test_labels, bestMod.predict(test_data), squared=True) # val
        
        nElements = np.sum(curr_solution)
        selected_elements_indices = np.where(curr_solution == 1)[0]
        vaScore[selected_elements_indices] = (vaScore[selected_elements_indices]+np.array([-accuracies[idx]]*nElements))*0.5
        print(-clf.best_score_, nElements, accuracies[idx])

        idx = idx + 1
    return (accuracies, vaScore)

def select_mating_pool(pop, fitness, num_parents):
    # Selecting the best individuals in the current generation as parents for producing the offspring of the next generation.
    parents = np.empty((num_parents, pop.shape[1]))
    for parent_num in range(num_parents):
        max_fitness_idx = np.where(fitness == np.max(fitness))
        max_fitness_idx = max_fitness_idx[0][0]
        parents[parent_num, :] = pop[max_fitness_idx, :]
        fitness[max_fitness_idx] = -99999999999
    return parents


def crossover(parents, offspring_size):
    offspring = np.empty(offspring_size)
    # The point at which crossover takes place between two parents. Usually, it is at the center.
    crossover_point = np.uint8(offspring_size[1]/2)

    for k in range(offspring_size[0]):
        # Index of the first parent to mate.
        parent1_idx = k%parents.shape[0]
        # Index of the second parent to mate.
        parent2_idx = (k+1)%parents.shape[0]
        # The new offspring will have its first half of its genes taken from the first parent.
        offspring[k, 0:crossover_point] = parents[parent1_idx, 0:crossover_point]
        # The new offspring will have its second half of its genes taken from the second parent.
        offspring[k, crossover_point:] = parents[parent2_idx, crossover_point:]
    return offspring


def mutation(offspring_crossover, varScore, score, num_mutations=2):
    fiftyArg = np.argwhere(score > np.percentile(score, 50))
    varInd = np.argwhere(varScore == 0)
    common = np.intersect1d(varInd, fiftyArg)
    if common.shape[0] != 0:
        print("left to test: ", common.shape[0])
        mutation_idx = np.random.randint(low=0, high=common.shape[0], size=num_mutations)
        mutation_idx = common[mutation_idx]
    else:
        oneArg = np.argwhere(offspring_crossover == 1)
        mutation_idx = np.random.randint(low=0, high=oneArg.shape[1], size=num_mutations)
        mutation_idx = oneArg[mutation_idx]
        #else:
         #   mutation_idx = np.random.randint(low=0, high=offspring_crossover.shape[1], size=num_mutations)
    for idx in range(offspring_crossover.shape[0]):
        offspring_crossover[idx, mutation_idx] = 1 - offspring_crossover[idx, mutation_idx]
    return offspring_crossover