#!/usr/bin/env python
# coding: utf-8

# In[50]:


import pandas as pd
import numpy as np
from functions import *
from scipy.signal import savgol_filter
from sklearn import model_selection

# set the columns that are interesting
colInterest = ["Weight (g)", "Top", "Bottom", "Core 1", "Core 2", "Sugar level", "pH level", "Salt", "avg Firmness", "avg Core", "avg Flesh"]
# read the csv file
dataSets = ['kiwiData_VNIR4.csv', 'kiwiData_SWIR4.csv', 'kiwiData_VNIR2.csv', 'kiwiData_SWIR2.csv']
 # has 1% removed
#Data = confRemoveList(Data, colInterest, 10) # 100-n confidence interval. keeps the colInterest

# In[51]:

import warnings

from sklearn.cross_decomposition import PLSRegression
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import mean_squared_error
import scipy


def preprocessing(doThis, myX):
    t = 10
    s = 0
    if len(doThis) > 3:  
        w, o, der, t, s = doThis
    else:
        w,o,der = doThis
    
    if t < 2:
        if t == 0:
            myX = SNV(myX)
        if t == 1:
            myX= msc(myX)
    if [w,o,der] != [-1,-1, -1]:
        myX = savgol_filter(myX, w, o, der, axis=1)

    if s == 1:
        StandardScaler().fit_transform(train)
    return myX
def rsquared(x, y):
    """ Return R^2 where x and y are array-like."""

    slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(x[:,0], y[:, 0])
    return r_value**2
def convertParam(params):
    
    
    res = ""
    t,s = -1, -1
    if len(params) > 3: 
        w, o, der, t, s = params
    else:
        w,o,der = params
        
    if t == 0:
        res = "SNV, "
    elif t == 1:
        res = "MSC, "
    if w != -1:
        res += "SavitzkyGolay("
        res += str(w)+","+str(o)+","+str(der)+")"
    else:
        return "N/A"
   
    return res
        
def testPLS(myX, myY, entire=False):
    if entire == False:
        parameters = {"n_components" : np.arange(1,np.int(myX.shape[1]*(1/3)))} # do a GridsearchCV for n_components in pls
    else:
        parameters = {"n_components" : np.arange(1,myX.shape[1])} # do a GridsearchCV for n_components in pls
    clf = GridSearchCV(PLSRegression(), parameters, cv=5, n_jobs=8, scoring='neg_root_mean_squared_error', verbose=0, refit=True)
    clf.fit(myX, myY)
    bestMod = clf.best_estimator_ # get best predictor
    
    return (-clf.best_score_, bestMod)
    #return (mean_squared_error(myYt, predVal, squared=True), bestMod)
# In[52]:

def testPLS2(myX, myY, myXt, myYt, entire=False):
    if entire == False:
        parameters = {"n_components" : np.arange(1,np.int(myX.shape[1]*(1/3)))} # do a GridsearchCV for n_components in pls
    else:
        parameters = {"n_components" : np.arange(1,myX.shape[1])} # do a GridsearchCV for n_components in pls
    clf = GridSearchCV(PLSRegression(), parameters, cv=5, n_jobs=8, scoring='neg_root_mean_squared_error', verbose=0, refit=True)
    clf.fit(myX, myY)
    bestMod = clf.best_estimator_ # get best predictor
    predVal = bestMod.predict(myXt)
    
    #return (-clf.best_score_, bestMod)
    return (mean_squared_error(myYt, predVal, squared=True), bestMod)
# In[52]:
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, MinMaxScaler

for fileName in dataSets:
    Data = pd.read_csv(fileName)
    colInterest = ["Weight (g)", "Top", "Bottom", "Core 1", "Core 2", "Sugar level", "pH level", "Salt", "avg Firmness", "avg Core", "avg Flesh"]
    Data = confRemoveList(Data, colInterest, 10)
    Feature = "avg Flesh" # feature to use

    #Data = Data[Data[Feature] < 20]
    myCols = ["Sugar level", "pH level", "avg Flesh"]
    myCols.reverse()
    for Feature in myCols:
        dt = Data.copy()
        X = dt["Spectra"].values.tolist()
        Y = dt[Feature].values.reshape(-1,1)

        res = [] # convert the "[1,1]" string to numpy array
        for z in X:
            spec = []
            for i in z.split():
                val = i.replace("[", "")
                val = val.replace("]", "")
                if val != "":
                    spec.append(np.float(val))
            res.append(spec)

        X = np.array(res)
        bestParam=[-1,-1,-1]
        
        trainIndices, remaining = spxy(X, Y, test_size=0.3)

        # use the spxy
        X_train, y_train = X[trainIndices], Y[trainIndices]
        indTest, rem = spxy(X[remaining], Y[remaining], test_size=0.5)
        X_val, y_val = X[remaining][rem], Y[remaining][rem]
        X_test, y_test = X[remaining][indTest], Y[remaining][indTest]

        
        from sklearn.neighbors import KNeighborsRegressor
        parameters = {"n_neighbors": np.arange(2,40)}
        clf = GridSearchCV(KNeighborsRegressor(), parameters, n_jobs=8, scoring='neg_root_mean_squared_error', verbose=0, refit=True)
        clf.fit(X_train, y_train)
        bestModKNN = [clf.best_estimator_]
        bestModKNN.append(bestModKNN[0].score(X_val, y_val))
        par = clf.best_params_
        bestModKNN.append(par[list(par.keys())[0]])
         
        
        import scipy

        # iteratively find the best spectra processing for given feature
        """
        plotdata = []
        bestParam = [-1,-1,-1]
        score = testPLS(X,Y)[0]
        for w in [i for i in range(10,17)]+[i for i in range(20,50)]: # loop through the salvgol filter parameters
            for o in range(2,7):
                for der in range(0,3):
                    for z in range(3): 
                        if o < w and w%2 == 1: # order has to be less than windows size. and w must be odd
                            train = preprocessing([w,o,der, z, 0], X)

                            thisScore= testPLS(train, Y, entire=False)[0]
                            print(thisScore, w, o, der, score, bestParam)
                            plotdata.append([w, thisScore, np.std(train, axis=1).mean()])
                            if thisScore < score: # is new best score
                                print("BEST     :)")
                                bestParam = [w,o,der]
                                score = thisScore
        plotdata = np.array(plotdata)
        np.savetxt('test.out', plotdata, delimiter=',')

        X = preprocessing(bestParam, X)
        """
        trainIndices, remaining = spxy(X, Y, test_size=0.3)

        # use the spxy
        X_train, y_train = X[trainIndices], Y[trainIndices]
        indTest, rem = spxy(X[remaining], Y[remaining], test_size=0.5)
        X_val, y_val = X[remaining][rem], Y[remaining][rem]
        X_test, y_test = X[remaining][indTest], Y[remaining][indTest]

        valInd = np.array(remaining)[rem]
        testInd = np.array(remaining)[indTest]
        
        from sklearn.neighbors import KNeighborsRegressor
        print(X_train.shape)
        parameters = {"n_neighbors": np.arange(2,40)}
        clf = GridSearchCV(KNeighborsRegressor(), parameters, n_jobs=8, scoring='neg_root_mean_squared_error', verbose=0, refit=True)
        clf.fit(X_train, y_train)
        bestModKNN2 = [clf.best_estimator_]
        bestModKNN2.append(bestModKNN2[0].score(X_val, y_val))
        par = clf.best_params_
        bestModKNN2.append(par[list(par.keys())[0]])
        
        sheet, camera, dtType = convert(Feature, fileName)
        print(sheet, camera, dtType)
        if bestModKNN2[1] > bestModKNN[1]:
            addRes("KNN.xlsx", sheet, dtType, camera, "KNN", convertParam(bestParam), bestModKNN2[2], bestModKNN2[0],[X_train, y_train], [X_test, y_test])
        else:
            addRes("KNN.xlsx", sheet, dtType, camera, "KNN", convertParam(bestParam), bestModKNN[2], bestModKNN[0],[X_train, y_train], [X_test, y_test])
        
        import statsmodels.api as sm
        from sklearn.cross_decomposition import PLSRegression
        from sklearn import model_selection
        import sklearn

        kf_10 = model_selection.KFold(n_splits=10, shuffle=True, random_state=1)
        size = X_train.shape[1]
        score = np.zeros(size) # make a list of scores for each variable
        maxScore = 5 # maximum score a model can give a variable
        #Backward Elimination variable selection
        cols = [i for i in range(X.shape[1])] # 0->186 list all variables
        pmax = 1
        rounds = 0 

        while (len(cols)>0):
            rounds += 1
            p= []
            X_1 = X[:, cols]

            model = sm.OLS(Y, X_1).fit()
            pmax = np.max(model.pvalues)
            if rounds == 1:
                score[cols] += maxScore*(1-model.pvalues/pmax) # use p_value as score for variables, only once

            feature_with_p_max = np.argmax(model.pvalues)
            if pmax > 0.05:
                print(feature_with_p_max, len(cols), pmax, np.min(model.pvalues), np.mean(model.pvalues))
                del cols[feature_with_p_max-1]
            else:
                break
        selected_features_BE = cols


        # In[21]:

        from sklearn.feature_selection import *
        from sklearn.linear_model import LinearRegression

        # variable scoring with variancethreshold and selectPercentile
        # make sure not the same variables are scored twice
        tempScore = np.zeros(size)

        sel = VarianceThreshold()
        sel.fit(X_train)
        score += sel.variances_

        tempScore = np.zeros(size)
        used = []
        for i in range(1,50):
            model = SelectPercentile(mutual_info_regression, percentile=i)
            try:  
                model.fit(X_train, y_train)
                mask = np.array(sel.get_support(indices=True))
                mask = mask[mask not in used]
                used.append(list(mask))
                tempScore[mask] += 1/i
            except:
                pass
        
        if np.sum(tempScore) != 0:
            score += (tempScore/np.max(tempScore))*maxScore

        import seaborn as sns
        from sklearn.feature_selection import *
        from sklearn.linear_model import LinearRegression

        # randomfrog (linear regression) variable scoring

        model = RFECV(PLSRegression(), step=1, cv=5, verbose=1, min_features_to_select=3)
        model.fit(X_train, y_train)
        z = model.ranking_ # get the rank of each variable, 1 is best
        score += maxScore/z

        # https://towardsdatascience.com/feature-reduction-using-genetic-algorithm-with-python-403a5f4ef0c1
        import numpy
        import GA
        import pickle
        import matplotlib.pyplot

        # genetic algorithm variable selection with pls (gridsearch cv)
        # kickstarted by the scoring done aboce

        data_inputs = X

        data_outputs = Y


        num_samples = data_inputs.shape[0]
        num_feature_elements = data_inputs.shape[1]

        train_indices = np.array(trainIndices)
        test_indices = np.array(valInd) # validation
        print("Number of training samples: ", train_indices.shape[0])
        print("Number of test samples: ", test_indices.shape[0])
        print(train_indices, test_indices)

        """
        Genetic algorithm parameters:
            Population size
            Mating pool size
            Number of mutations
        """
        sol_per_pop = 16 # Population size.
        num_parents_mating = 8 # Number of parents inside the mating pool.
        num_mutations = 3 # Number of elements to mutate.

        # Defining the population shape.
        pop_shape = (sol_per_pop, num_feature_elements)

        # Creating the initial population.
        new_population = numpy.random.randint(low=0, high=2, size=pop_shape)
        print(new_population.shape)

        # kickstart the genetic algorithm
        start = 90 
        end = 99 # 50-80
        index = 0
        for i in range(sol_per_pop):
            t = start+int(((end-start))/sol_per_pop)*i
            count = np.sum(np.where(score > np.percentile(score, t), 1, 0))
            print(t, count)
            if count < 100 and count > 1:
                new_population[index] = np.where(score > np.percentile(score, t), 1, 0)
                index += 1
        if sheet == "Firmness" and camera == "VNIR":
            new_population[0] = [0,0,1,0,1,1,0,0,1,0,0,0,1,0,1,1,0,1,1,0,0,1,1,1,0,0,0,0,1,0,0,0,0,0,1,0,1,0,0,0,0,1,0,1,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,1,0,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,0,0,1,1,0,1,1,0,1,0,1,0,0,0,1,0,1,1,0,1,0,1,1,0]
            new_population[1] = [0,0,1,0,0,1,1,0,1,0,0,1,1,0,1,1,1,1,0,0,0,1,1,1,0,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,0,1,0,0,0,0,0,0,1,0,1,1,1,1,1,0,0,0,1,1,0,0,1,1,0,0,1,0,0,0,0,0,0,1,0,0,1,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,1,0,1,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,1,1,0,0,0,0,0,0,0,1,1,0,1,1,0,1,0,1,0,0,0,1,0,1,1,0,1,0,1,1,1]
        
        best_outputs = []
        varScore = np.zeros(X_train.shape[1])
        num_generations = 100
        for generation in range(num_generations):
            print("Generation : ", generation)
            # Measuring the fitness of each chromosome in the population.
            fitness, tempScore = GA.cal_pop_fitness(new_population, data_inputs, data_outputs, train_indices, test_indices)
            varScore = (tempScore+varScore)*0.5
            best_outputs.append(numpy.max(fitness))
            # The best result in the current iteration.
            print("Best result : ", best_outputs[-1], Feature, fileName)

            # Selecting the best parents in the population for mating.
            parents = GA.select_mating_pool(new_population, fitness, num_parents_mating)

            # Generating next generation using crossover.
            offspring_crossover = GA.crossover(parents, offspring_size=(pop_shape[0]-parents.shape[0], num_feature_elements))

            # Adding some variations to the offspring using mutation.
            offspring_mutation = GA.mutation(offspring_crossover, varScore, score, num_mutations=num_mutations)

            # Creating the new population based on the parents and offspring.
            new_population[0:parents.shape[0], :] = parents
            new_population[parents.shape[0]:, :] = offspring_mutation

        # Getting the best solution after iterating finishing all generations.
        # At first, the fitness is calculated for each solution in the final generation.
        fitness, bestEst = GA.cal_pop_fitness(new_population, data_inputs, data_outputs, train_indices, test_indices)
        # Then return the index of that solution corresponding to the best fitness.
        best_match_idx = numpy.where(fitness == numpy.max(fitness))[0]
        best_match_idx = best_match_idx[0]

        best_solution = new_population[best_match_idx, :]
        best_solution_indices = numpy.where(best_solution == 1)[0]
        best_solution_num_elements = best_solution_indices.shape[0]
        best_solution_fitness = fitness[best_match_idx]

        print("best_match_idx : ", best_match_idx)
        print("best_solution : ", best_solution)
        print("Selected indices : ", best_solution_indices)
        print("Number of selected elements : ", best_solution_num_elements)
        print("Best solution fitness : ", best_solution_fitness)
        
        np.savetxt(sheet+camera+dtType, best_solution_indices, delimiter=',')
        np.savetxt(sheet+camera+dtType+"score", varScore, delimiter=',')
        
        X_train, X_test, X_val = X_train[:, best_solution], X_test[:, best_solution], X_val[:, best_solution]
        model = testPLS2(X_train, y_train, X_val, y_val, entire=True)[1]
        addRes("GAPLS.xlsx", sheet, dtType, camera, "GAPLS", convertParam(bestParam), best_solution_num_elements, model, (X_train, y_train), (X_test, y_test))

# In[ ]:







# In[ ]:
