#!/usr/bin/env python
# coding: utf-8

# In[50]:


import pandas as pd
import numpy as np
from functions import *
from scipy.signal import savgol_filter
from sklearn import model_selection
from scipy.optimize import differential_evolution

# set the columns that are interesting
colInterest = ["Weight (g)", "Top", "Bottom", "Core 1", "Core 2", "Sugar level", "pH level", "Salt", "avg Firmness", "avg Core", "avg Flesh"]
# read the csv file
dataSets = ['kiwiData_VNIR4.csv','kiwiData_SWIR4.csv', 'kiwiData_SWIR2.csv', 'kiwiData_VNIR2.csv']
 # has 1% removed
#Data = confRemoveList(Data, colInterest, 10) # 100-n confidence interval. keeps the colInterest

# In[51]:


from sklearn.cross_decomposition import PLSRegression
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import mean_squared_error
import scipy
from sklearn.decomposition import KernelPCA
from sklearn.neighbors import KNeighborsRegressor
from sklearn.model_selection import PredefinedSplit

def preprocessing(doThis, myX):
    t = 10
    s = 0
    n = -1
    if len(doThis) == 5:  
        w, o, der, t, s = doThis
    elif len(dothis) == 4:
        w, o, der, t = doThis
    else:
        w,o,der = doThis
    
    if t < 2:
        if t == 0:
            myX = SNV(myX)
        if t == 1:
            myX= msc(myX)
    if w != -1 and o != -1 and der != -1:
        print(w,o,der)
        myX = savgol_filter(myX, w, o, der, axis=1)

    if s == 1:
        StandardScaler().fit_transform(myX)
    return myX
def rsquared(x, y):
    """ Return R^2 where x and y are array-like."""

    slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(x[:,0], y[:, 0])
    return r_value**2
def convertParam(params):
    
    
    res = ""
    t,s = -1, -1
    if len(params) > 3: 
        w, o, der, t, s = params
    else:
        w,o,der = params
        
    if t == 0:
        res = "SNV, "
    elif t == 1:
        res = "MSC, "
    if w != -1:
        res += "SavitzkyGolay("
        res += str(w)+","+str(o)+","+str(der)+")"
    else:
        return "N/A"
   
    return res
        
def testPLS(myX, myY, myXt, myYt, entire=False):
    if entire == False:
        parameters = {"n_components" : np.arange(1,np.int(myX.shape[1]*(1/3)))} # do a GridsearchCV for n_components in pls
    else:
        parameters = {"n_components" : np.arange(1,myX.shape[1])} # do a GridsearchCV for n_components in pls
    clf = GridSearchCV(PLSRegression(), parameters, cv=5, n_jobs=8, scoring='neg_root_mean_squared_error', verbose=0, refit=True)
    clf.fit(myX, myY)
    bestMod = clf.best_estimator_ # get best predictor
    predVal = bestMod.predict(myXt)
    
    #return (-clf.best_score_, bestMod)
    return (mean_squared_error(myYt, predVal, squared=True), bestMod)
# In[52]:
iterations = 0
def minimize(x):
    global iterations, X, Y,score,bestParam
    w = int(x[0])
    o = int(x[1])
    der = int(x[2])
    z = int(x[3])
    n = int(x[4])
    if o < w and w%2 == 1:
        train = X
        if iterations != 0:
            train = preprocessing([w,o,der, z, 0], X)

        train = StandardScaler().fit_transform(train)
        trainIndices, remaining = spxy(train, Y, test_size=0.3)
        X_train, y_train = train[trainIndices], Y[trainIndices]
        indTest, rem = spxy(train[remaining], Y[remaining], test_size=0.5)
        X_val, y_val = train[remaining][rem], Y[remaining][rem]
        X_test, y_test = train[remaining][indTest], Y[remaining][indTest]

        valInd = np.array(remaining)[rem]
        testInd = np.array(remaining)[indTest]
        if n != -1:
            train = KernelPCA(n_components=n, kernel='rbf').fit_transform(train)
            X_train, X_val, X_test = KernelPCA(n_components=n, kernel='rbf').fit_transform(X_train), KernelPCA(n_components=n, kernel='rbf').fit_transform(X_val), KernelPCA(n_components=n, kernel='rbf').fit_transform(X_test)
        Xvaltrain = []
        Yvaltrain = []
        mask = []
        for i in list(valInd)+list(trainIndices):
            Xvaltrain.append(train[i, :])
            Yvaltrain.append(Y[i, :])
            if i in valInd:
                mask.append(0)
            else:
                mask.append(-1)
        Xvaltrain = np.array(Xvaltrain)
        print(Xvaltrain.shape, X_train.shape)
        pds = PredefinedSplit(test_fold = mask)
        parameters = {"n_neighbors": np.arange(2,X_train.shape[0]-100)}

        clf = GridSearchCV(KNeighborsRegressor(), parameters, n_jobs=8, cv=pds,scoring='neg_root_mean_squared_error', verbose=0, refit=True)
        clf.fit(Xvaltrain, Yvaltrain)
        par = clf.best_params_
        model = clf.best_estimator_
        thisScore = clf.best_score_
        print(thisScore, w, o, der, z,n, score, bestParam, X_train.mean())
        iterations += 1
        if thisScore > score: # is new best score
            print("BEST     :)")
            bestParam = [w,o,der, z, 0]
            score = thisScore
            addRes("KNN.xlsx", sheet, dtType, camera, "KNN", convertParam(bestParam), par[list(par.keys())[0]], model,[X_train, y_train], [X_test, y_test])
        return -thisScore
    return 10000

from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, MinMaxScaler

for fileName in dataSets:
    Data = pd.read_csv(fileName)
    
    Feature = "avg Flesh" # feature to use

    #Data = Data[Data[Feature] < 20]
    #myCols = ["Sugar level", "pH level", "avg Flesh"]
    myCols = ["Salt"]
    #myCols.reverse()
    for Feature in myCols:
        X = Data["Spectra"].values.tolist()
        Y = Data[Feature].values.reshape(-1,1)

        res = [] # convert the "[1,1]" string to numpy array
        for z in X:
            spec = []
            for i in z.split():
                val = i.replace("[", "")
                val = val.replace("]", "")
                if val != "":
                    spec.append(np.float(val))
            res.append(spec)

        X = np.array(res)
        X = StandardScaler().fit_transform(X)
        Y = StandardScaler().fit_transform(Y)
        

        
        """
        bestModKNN = [0,-1000]
        ncomp = 0
        for n in range(2, X_train.shape[0]):
            trans = KernelPCA(n_components=n, kernel='rbf')
            train = trans.fit_transform(X)

            print(X_train.shape)
            parameters = {"n_neighbors": np.arange(2,train.shape[0])}
            clf = GridSearchCV(KNeighborsRegressor(), parameters, n_jobs=8, scoring='neg_root_mean_squared_error', verbose=0, refit=True)
            clf.fit(X, Y)
            if clf.best_score_ > bestKNN[1]:
                ncomp = n
                bestModKNN = [clf.best_estimator_]
                bestModKNN.append(clf.best_score_ )
                par = clf.best_params_
                bestModKNN.append(par[list(par.keys())[0]])
        """ 
        import scipy
        sheet, camera, dtType = convert(Feature, fileName)
        # iteratively find the best spectra processing for given feature
        
        plotdata = []
        bestParam = [-1,-1,-1]
        score = -1000
        iterations = 0
        bounds = [(20, 50), (2,7), (0,3),(0,3),(9,60)]
        result = differential_evolution(minimize, bounds, maxiter=100)
                                