#!/usr/bin/env python
# coding: utf-8

# # Imports and reference reflactance

# In[1]:


import numpy as np              # for math and arrays
import spectral as sp           # for spectral image reading
import matplotlib.pyplot as plt # for plots and figures
import scipy.ndimage as ndimage # for medianFilter
from functions import *         # helping functions
import os                       # for checking folder existance
import cv2                      # for meanFilter
import time                     # for timing time usage
from scipy import signal

lines = []  # for txt file reading
arr = []    # for 2-dim array with reference values + wavelength
with open("SRT-MS-050.txt", "r") as file:  # open file
    lines = file.readlines()               # read all lines
    for line in lines:                     # loop through each line
        for number in line.split():        # loop through each number in line
            try:                           # if not a number, skip
                arr.append(float(number))
            except ValueError:
                pass
file.close()

# make 2-dim array of all values for reference target
temp_arr = np.round((np.array(arr).reshape((2251, 5))), 4)

# copy last values into 3 more rows
arr = np.zeros((2275, 5))
arr[:2251] = temp_arr[:2251]
arr[-(2275-2251):] = temp_arr[-1]


# # Image names and reference target

# In[2]:
def meanFilter(img, kSize):
    kernel = np.ones((kSize, kSize), np.float32)/(kSize*kSize)
    return cv2.filter2D(img, -1, kernel)
def smooth(x,window_len=11,window='hanning'):
    if x.ndim != 1:
        print("smooth only accepts 1 dimension arrays.")
        return 0

    if x.size < window_len:
        print("Input vector needs to be bigger than window size.")
        return 0

    if window_len<3:
        return x

    s=np.r_[x[window_len-1:0:-1],x,x[-2:-window_len-1:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w=np.ones(window_len,'d')
    else:
        w=eval('np.'+window+'(window_len)')

    y=np.convolve(w/w.sum(),s,mode='valid')
    return y[window_len//2:-window_len//2+1]
    
def normalize(img_name, arr, ref_percent, ref_y0, ref_y1, kSize, kSizeHann):
    img = sp.open_image(img_name) #[:600,:,:] #for testing on a smaller subimage
    wavelengths = img.bands.centers
    bands = len(wavelengths)
    ref = meanXcoord(img[ref_y0:ref_y1, :])

    img = img[ref_y1:,:] # crops the image a bit
    #img = img[:,:]

    light = np.zeros((img.shape[1], img.shape[2]))

    for band in range(bands):
        for x in range(img.shape[1]):
            light[x, band] = ref[x, band] / arr[ np.int(wavelengths[band]) - 250, ref_percent]
        

    # Finding reflectance
    img_normalized = np.zeros(img.shape, dtype=np.float32)
    print("img_normalized shape = ", img_normalized.shape)
    
    print("Normalisation starting!")

    for b in range(bands):
        for x in range(img.shape[1]):
            img_normalized[:, x, b] = img[:, x, b] / light[x, b]
        #img_normalized[:, :, b] = ndimage.median_filter(img_normalized[:, :, b], size=(kSize,kSize))
        if b % 15 == 0:
            print("Status: %.2f " % ( np.float(b/bands)*100), "percent done" ) # 1 = 100% ferdig
    
    #print("Smoooothing")
    #for y in range(img.shape[0]):
    #    for x in range(img.shape[1]):
    #        img_normalized[y, x, :] = smooth(img_normalized[y, x, :], kSizeHann)
    #    if y % 200 == 0:
    #        print("Status: %.2f " % ( np.float(y/img.shape[0])*100), " done")
            
            
    return img_normalized


# # Normalization


srcFoldr = "E:/kiwi/3/day2/rad"
normFoldr = "E:/kiwi/3/day2/norm"

#srcFoldr = "E:/kiwi/colorcheck/rad"
#normFoldr = "E:/kiwi/colorcheck/norm"

if not os.path.exists(srcFoldr) and not os.path.exists(normFoldr):
    print("Source folder or normFolder does not exist")
    exit()

ref_percent = 4 # | 1: 99% | 2: 50% | 3: 25% | 4: 10%
y0Swir = 380
y0Nir = 1760

vnir_names, swir_names = getFiles(srcFoldr, normFoldr)
allFiles = vnir_names+swir_names
allFiles = ["E:/kiwi/2/day2/rad/test2_day2_box5_b1_VNIR_1800_SN00841_6059us_2021-03-03T151147_raw_rad.hdr"]
for i, file in enumerate(allFiles):
    print("Normaliseing: "+ file)
    print("This is image number ", allFiles.index(file)+1," out of ", len(allFiles), "Images")
    print(np.float(i/len(allFiles))*100, "percent")
    
    timeNow = time.time()
    if "VNIR" in file:
        img_norm = normalize(file, arr, ref_percent, y0Nir, y0Nir+40, 3, 11)
    elif "SWIR" in file:
        img_norm = normalize(file, arr, ref_percent, y0Swir, y0Swir+40, 3, 11)
    
    print("Time used on norm (S): ", time.time()-timeNow)
    
    saveName = file.replace("rad", "testing") # swap the srcfolder with normalise folder
    saveName = saveName.replace(".hdr", "_norm.hdr") # put in norm ending of file
    print("Saving image: ", saveName)
    sp.envi.save_image(saveName, img_norm, force=True)
    print(i+1," out of ", len(allFiles), "images finished")


    

