#!/usr/bin/env python
# coding: utf-8

# # Imports and reference reflactance

# In[1]:


import numpy as np              # for math and arrays
import spectral as sp           # for spectral image reading
import os
import cv2
import matplotlib.pyplot as plt # for plots and figures
from functions import *         # helping functions
import time

lines = []  # for txt file reading
arr = []    # for 2-dim array with reference values + wavelength
with open("SRT-MS-050.txt", "r") as file:  # open file
    lines = file.readlines()               # read all lines
    for line in lines:                     # loop through each line
        for number in line.split():        # loop through each number in line
            try:                           # if not a number, skip
                arr.append(float(number))
            except ValueError:
                pass
file.close()

# make 2-dim array of all values for reference target
temp_arr = np.round((np.array(arr).reshape((2251, 5))), 4)

# copy last values into 3 more rows
arr = np.zeros((2266, 5))
arr[:2251] = temp_arr[:2251]
arr[-15:] = temp_arr[-1]


# # Image names and reference target

# In[2]:
srcFoldr = "F:/1/day2/rad"
normFoldr = "F:/1/day2/norm"
#while not os.path.exists(srcFoldr) and not os.path.exists(normFoldr):
#    srcFoldr = input("source folder: ")
#    normFoldr = input("norm folder: ")
vnir_names, swir_names = getFiles(srcFoldr, normFoldr)



# y1 63 y2 183


# In[3]:
def meanFilter(img, kSize):
    kernel = np.ones((kSize, kSize), np.float32)/(kSize*kSize)
    return cv2.filter2D(img, -1, kernel)

def normalize(img_name, arr, ref_percent, ref_y0, ref_y1):
    img = sp.open_image(img_name)#[:600,:,:] #for testing on a smaller subimage
    
    ref = meanXcoord(img[ref_y0:ref_y1, :])
    
    # hvis du vil se hvor på bildet ref_y0:ref_y1 går:
        #test = np.copy(img)
        #test[ref_y0:ref_y1, :] = 0
        #plt.imshow(test[:,:,170])
        #plt.title("Reference selection")
        #plt.figure()

    img = img[ref_y1:,:] # crops the image a bit
    

    # Calvulate wavelength increments and wavelength start/stop
    # IF VNIR
    if (img_name.find("VNIR")):
        start=157; end=start+590; increment=3.1754; bands=186
    # IF SWIR
    elif (img_name.find("SWIR")):
        start=701; end=start+1565; increment=5.434; bands=288
        # (951,66 - 2516,66) -> 1565
    # IF none
    else:
        print("Could not find VNIR or SWIR in filename: ", img_name)
        return 0
    
    # Get light values for each band
    band = 0
    light = np.zeros((img.shape[1], img.shape[2]))
    
    for i in np.arange(start, end, increment):
        for x in range(img.shape[1]):
            light[x, band] = ref[x, band] / arr[np.int(i), ref_percent]
        band += 1
        

    # Finding reflectance
    img_normalized = np.zeros((img.shape[0], img.shape[1], img.shape[2]))
    print("Normalisation startet!")

    for b in range(bands):
        
        for x in range(img.shape[1]):
            img_normalized[:, x, b] = img[:, x, b] / light[x, b]
        
        img_normalized[:, :, b] = meanFilter(img_normalized[:, :, b], 3)
        
        
        
    print("Status: %f " % ( np.float(b/bands)*100), "percent done" ) # 1 = 100% ferdig
    
    
    return img_normalized


# # Normalization

# In[5]:


# uncomment for user input
#ref_percent = int(input("ref precent | 1: 99% | 2: 50% | 3: 25% | 4: 10%: ")) # | 1: 99% | 2: 50% | 3: 25% | 4: 10%
#y0Swir = int(input("y number swir: "))
#y0Nir = int(input("y number vnir: "))
ref_percent = 4
y0Swir = 390
y0Nir = 1780



allFiles = vnir_names+swir_names
for i, file in enumerate(allFiles):
    print("Normaliseing: "+ file)
    print("This is image number ", allFiles.index(file)+1," out of ", len(allFiles), "Images")
    print(np.float(i/len(allFiles))*100, "percent")
    
    timeNow = time.time()
    if "VNIR" in file:
        img_norm = normalize(file, arr, ref_percent, y0Nir, y0Nir+40)
    elif "SWIR" in file:
        img_norm = normalize(file, arr, ref_percent, y0Swir, y0Swir+40)
    
    print("time used on norm (ms): ", time.time()-timeNow)
    print("max and min", img_norm.max(), img_norm.min())
    
    saveName = file.replace(srcFoldr, normFoldr) # swap the srcfolder with normalise folder
    saveName = saveName.replace(".hdr", "_norm.hdr") # put in norm ending of file
    
    sp.envi.save_image(saveName, img_norm)
    print(i+1," out of ", len(allFiles), "Images finished")    


# In[ ]:




